import { CompilerEvent, CompilerCallbackOptions, CompilerMessageOverrideMap } from "./compiler-interfaces.js";
/**
 * Abstract interface for callbacks, to abstract out file i/o
 */
export interface CompilerCallbacks {
    /**
     * Attempt to load a file. Return falsy if not found.
     * @param filename
     */
    loadFile(filename: string): Uint8Array;
    /**
     * Get file size, returns undefined if not found
     */
    fileSize(filename: string): number;
    /**
     * Returns true if file is a directory, undefined if not found
     */
    isDirectory(filename: string): boolean;
    get path(): CompilerPathCallbacks;
    get fs(): CompilerFileSystemCallbacks;
    get fsAsync(): CompilerFileSystemAsyncCallbacks;
    get net(): CompilerNetAsyncCallbacks;
    /**
     * Resolves a file path relative to the baseFilename
     * @param baseFilename
     * @param filename
     */
    resolveFilename(baseFilename: string, filename: string): string;
    reportMessage(event: CompilerEvent): void;
    debug(msg: string): void;
    fileURLToPath(url: string | URL): string;
}
/**
 * A mapping for common path operations, maps to Node path module. This only
 * defines the functions we are actually using, so that we can port more easily
 * between different systems.
 */
export interface CompilerPathCallbacks {
    dirname(name: string): string;
    extname(name: string): string;
    basename(name: string, ext?: string): string;
    isAbsolute(name: string): boolean;
    join(...paths: string[]): string;
    normalize(p: string): string;
    relative(from: string, to: string): string;
    resolve(...args: string[]): string;
}
/**
 * A mapping for common filesystem operations, maps to Node fs module. This only
 * defines the functions we are actually using, so that we can port more easily
 * between different systems.
 */
export interface CompilerFileSystemCallbacks {
    readdirSync(name: string): string[];
    readFileSync(path: string, options?: {
        encoding?: null;
        flag?: string;
    } | null): Uint8Array;
    readFileSync(path: string, options: {
        encoding: string;
        flag?: string;
    } | string): string;
    readFileSync(path: string, options?: {
        encoding?: string | null;
        flag?: string;
    } | string | null): string | Uint8Array;
    writeFileSync(path: string, data: Uint8Array): void;
    mkdirSync(path: string, options?: {
        recursive?: boolean;
    }): string;
    existsSync(name: string): boolean;
}
export interface CompilerAsyncCallbacks {
    readonly fsAsync: CompilerFileSystemAsyncCallbacks;
    readonly path: CompilerPathCallbacks;
    readonly net: CompilerNetAsyncCallbacks;
}
export interface CompilerFileSystemCallbacksFolderEntry {
    filename: string;
    type: 'file' | 'dir';
}
export interface CompilerNetAsyncCallbacks {
    fetchJSON(url: string): Promise<any>;
    fetchBlob(url: string): Promise<Uint8Array>;
}
export interface CompilerFileSystemAsyncCallbacks {
    readFile(filename: string): Promise<Uint8Array>;
    readdir(filename: string): Promise<CompilerFileSystemCallbacksFolderEntry[]>;
    exists(filename: string): Promise<boolean>;
    resolveFilename(baseFilename: string, filename: string): string;
}
/**
 * Wrapper class for CompilerCallbacks for a given input file
 */
export declare class CompilerFileCallbacks implements CompilerCallbacks {
    private filename;
    private options;
    private parent;
    messages: CompilerEvent[];
    constructor(filename: string, options: CompilerCallbackOptions, parent: CompilerCallbacks);
    /**
     * Returns `true` if any message in the `messages` array is a Fatal or Error
     * message, and if `compilerWarningsAsErrors` is `true`, then also returns
     * `true` if any message is a Warning.
     */
    static hasFailureMessage(messages: CompilerEvent[], compilerWarningsAsErrors: boolean): boolean;
    /**
     *
     * @param event
     * @param overrides
     * @returns true if event has been suppressed
     */
    static applyMessageOverridesToEvent(event: CompilerEvent, overrides: CompilerMessageOverrideMap): boolean;
    /**
     * Returns `true` if any message in the `messages` array is a Fatal or Error
     * message, and if `compilerWarningsAsErrors` is `true`, then also returns
     * `true` if any message is a Warning.
     *
     * If passed a defined `compilerWarningsAsErrors` value, then uses that,
     * otherwise uses `options.compilerWarningsAsErrors`, or `false` if that is
     * also `undefined`.
     */
    hasFailureMessage(compilerWarningsAsErrors?: boolean): boolean;
    clear(): void;
    loadFile(filename: string): Uint8Array;
    fileSize(filename: string): number;
    isDirectory(filename: string): boolean;
    get path(): CompilerPathCallbacks;
    get fs(): CompilerFileSystemCallbacks;
    get net(): CompilerNetAsyncCallbacks;
    get fsAsync(): CompilerFileSystemAsyncCallbacks;
    resolveFilename(baseFilename: string, filename: string): string;
    reportMessage(event: CompilerEvent): void;
    debug(msg: string): void;
    fileURLToPath(url: string | URL): string;
}
export declare class DefaultCompilerFileSystemAsyncCallbacks implements CompilerFileSystemAsyncCallbacks {
    private owner;
    constructor(owner: CompilerCallbacks);
    exists(filename: string): Promise<boolean>;
    readFile(filename: string): Promise<Uint8Array>;
    readdir(filename: string): Promise<CompilerFileSystemCallbacksFolderEntry[]>;
    resolveFilename(baseFilename: string, filename: string): string;
}
//# sourceMappingURL=compiler-callbacks.d.ts.map