/*
 * Keyman is copyright (C) SIL Global. MIT License.
 *
 * Callbacks for Keyman Developer compilers to ensure independence from
 * filesystem, network, and console.
 */
import { CompilerErrorSeverity, CompilerError, CompilerErrorMask } from "./compiler-interfaces.js";
;
/**
 * Wrapper class for CompilerCallbacks for a given input file
 */
export class CompilerFileCallbacks {
    filename;
    options;
    parent;
    messages = [];
    constructor(filename, options, parent) {
        this.filename = filename;
        this.options = options;
        this.parent = parent;
    }
    /**
     * Returns `true` if any message in the `messages` array is a Fatal or Error
     * message, and if `compilerWarningsAsErrors` is `true`, then also returns
     * `true` if any message is a Warning.
     */
    static hasFailureMessage(messages, compilerWarningsAsErrors) {
        const failureCodes = [
            CompilerErrorSeverity.Fatal, CompilerErrorSeverity.Error
        ].concat(compilerWarningsAsErrors ? [CompilerErrorSeverity.Warn] : []);
        return messages.find(m => failureCodes.includes(CompilerError.severity(m.code))) != undefined;
    }
    /**
     *
     * @param event
     * @param overrides
     * @returns true if event has been suppressed
     */
    static applyMessageOverridesToEvent(event, overrides) {
        // Override event severity from user preference -- this will not override
        // fatal or error events
        const severity = overrides?.[CompilerError.error(event.code)] ??
            CompilerError.severity(event.code);
        if (severity == 'disable') {
            return true;
        }
        // Override the default event severity with the command line option
        event.code = severity | (event.code & ~CompilerErrorMask.Severity);
        return false;
    }
    /**
     * Returns `true` if any message in the `messages` array is a Fatal or Error
     * message, and if `compilerWarningsAsErrors` is `true`, then also returns
     * `true` if any message is a Warning.
     *
     * If passed a defined `compilerWarningsAsErrors` value, then uses that,
     * otherwise uses `options.compilerWarningsAsErrors`, or `false` if that is
     * also `undefined`.
     */
    hasFailureMessage(compilerWarningsAsErrors) {
        return CompilerFileCallbacks.hasFailureMessage(this.messages, compilerWarningsAsErrors ?? this.options.compilerWarningsAsErrors ?? false);
    }
    clear() {
        this.messages = [];
    }
    loadFile(filename) {
        return this.parent.loadFile(filename);
    }
    fileSize(filename) {
        return this.parent.fileSize(filename);
    }
    isDirectory(filename) {
        return this.parent.isDirectory(filename);
    }
    get path() {
        return this.parent.path;
    }
    get fs() {
        return this.parent.fs;
    }
    get net() {
        return this.parent.net;
    }
    get fsAsync() {
        return this.parent.fsAsync;
    }
    resolveFilename(baseFilename, filename) {
        return this.parent.resolveFilename(baseFilename, filename);
    }
    reportMessage(event) {
        const disable = CompilerFileCallbacks.applyMessageOverridesToEvent(event, this.options.messageOverrides);
        this.messages.push(event);
        if (!disable) {
            this.parent.reportMessage({ filename: this.filename, ...event });
        }
    }
    debug(msg) {
        return this.parent.debug(msg);
    }
    fileURLToPath(url) {
        return this.parent.fileURLToPath(url);
    }
}
export class DefaultCompilerFileSystemAsyncCallbacks {
    owner;
    constructor(owner) {
        this.owner = owner;
    }
    async exists(filename) {
        return this.owner.fs.existsSync(filename);
    }
    async readFile(filename) {
        return this.owner.loadFile(filename);
    }
    async readdir(filename) {
        return this.owner.fs.readdirSync(filename).map(item => ({
            filename: item,
            type: this.owner.isDirectory(this.owner.path.join(filename, item)) ? 'dir' : 'file'
        }));
    }
    resolveFilename(baseFilename, filename) {
        return this.owner.resolveFilename(baseFilename, filename);
    }
}
//# sourceMappingURL=compiler-callbacks.js.map