//
// Version 1.0 and 2.0 of Keyman Developer Project .kpj file
//
import { KeymanFileTypes } from '@keymanapp/common-types';
export class KeymanDeveloperProject {
    _projectFilename;
    callbacks;
    options;
    files;
    projectPath = '';
    projectFile;
    get projectFilename() {
        return this._projectFilename;
    }
    constructor(_projectFilename, version, callbacks) {
        this._projectFilename = _projectFilename;
        this.callbacks = callbacks;
        this.projectPath = this.callbacks.path.dirname(this._projectFilename);
        this.options = new KeymanDeveloperProjectOptions(version);
        this.files = [];
        this.projectFile = new KeymanDeveloperProjectFile20(_projectFilename, this.callbacks.path);
    }
    /**
     * Adds .kmn, .xml, .kps to project based on options.sourcePath
     * @param projectFilename Full path to project.kpj (even if the file doesn't exist)
     */
    async populateFiles() {
        if (this.options.version != '2.0') {
            throw new Error('populateFiles can only be called on a v2.0 project');
        }
        const sourcePath = this.resolveProjectPath(this.options.sourcePath);
        if (!await this.callbacks.fsAsync.exists(sourcePath)) {
            return false;
        }
        const files = await this.callbacks.fsAsync.readdir(sourcePath);
        for (const file of files.filter(file => file.type == 'file')) {
            const fullPath = this.callbacks.path.join(sourcePath, file.filename);
            if (KeymanFileTypes.filenameIs(file.filename, ".xml" /* KeymanFileTypes.Source.LdmlKeyboard */)) {
                try {
                    const data = await this.callbacks.fsAsync.readFile(fullPath);
                    const text = new TextDecoder().decode(data);
                    if (!text?.match(/<keyboard3/)) {
                        // Skip this .xml because we assume it isn't really a keyboard .xml
                        continue;
                    }
                }
                catch (e) {
                    // We'll just silently skip this file because we were not able to load it,
                    // so let's hope it wasn't a real LDML keyboard XML :-)
                    continue;
                }
            }
            if (KeymanFileTypes.sourceTypeFromFilename(file.filename) !== null) {
                const file = new KeymanDeveloperProjectFile20(fullPath, this.callbacks.path);
                this.files.push(file);
            }
        }
        return this.files.length > 0;
    }
    isKeyboardProject() {
        return !!this.files.find(file => file.fileType == ".kmn" /* KeymanFileTypes.Source.KeymanKeyboard */ || file.fileType == ".xml" /* KeymanFileTypes.Source.LdmlKeyboard */);
    }
    isLexicalModelProject() {
        return !!this.files.find(file => file.fileType == ".model.ts" /* KeymanFileTypes.Source.Model */);
    }
    resolveProjectPath(p) {
        // Replace placeholders in the target path
        return p.replace('$PROJECTPATH', this.projectPath);
    }
    resolveBuildPath() {
        // Roughly corresponds to Delphi TProject.GetTargetFileName
        let p = this.options.version == '1.0' ?
            this.options.buildPath || '$SOURCEPATH' :
            this.options.buildPath;
        // Replace placeholders in the target path
        if (this.options.version == '1.0') {
            // TODO: do we need to support $VERSION?
            // $SOURCEPATH only supported in 1.0 projects
            p = p.replace('$SOURCEPATH', this.callbacks.path.dirname(this._projectFilename));
        }
        p = this.resolveProjectPath(p);
        return p;
    }
    getOutputFilePath(type) {
        const p = this.resolveBuildPath();
        const f = this.callbacks.path.basename(this._projectFilename, ".kpj" /* KeymanFileTypes.Source.Project */) + type;
        return this.callbacks.path.normalize(this.callbacks.path.join(p, f));
    }
    resolveInputFilePath(file) {
        return this.callbacks.fsAsync.resolveFilename(this._projectFilename, file.filePath);
    }
    resolveOutputFilePath(file, sourceExt, targetExt) {
        // Roughly corresponds to Delphi TProject.GetTargetFileName
        let p = this.options.version == '1.0' ?
            this.options.buildPath || '$SOURCEPATH' :
            this.options.buildPath;
        // Replace placeholders in the target path
        if (this.options.version == '1.0') {
            // TODO: do we need to support $VERSION?
            // $SOURCEPATH only supported in 1.0 projects
            p = p.replace('$SOURCEPATH', this.callbacks.path.dirname(this.resolveInputFilePath(file)));
        }
        p = this.resolveProjectPath(p);
        const f = file.filename.replace(new RegExp(`\\${sourceExt}$`, 'i'), targetExt);
        return this.callbacks.path.normalize(this.callbacks.path.join(p, f));
    }
}
;
export var KeymanDeveloperProjectType;
(function (KeymanDeveloperProjectType) {
    KeymanDeveloperProjectType[KeymanDeveloperProjectType["Keyboard"] = 0] = "Keyboard";
    KeymanDeveloperProjectType[KeymanDeveloperProjectType["LexicalModel"] = 1] = "LexicalModel";
})(KeymanDeveloperProjectType || (KeymanDeveloperProjectType = {}));
;
export class KeymanDeveloperProjectOptions {
    buildPath;
    sourcePath;
    compilerWarningsAsErrors = false;
    warnDeprecatedCode = true;
    checkFilenameConventions = false; // missing option defaults to False
    /**
     * Skip building .keyboard_info and .model_info files, for example in
     * unit tests or for legacy keyboards
     */
    skipMetadataFiles;
    projectType = KeymanDeveloperProjectType.Keyboard;
    version;
    constructor(version) {
        this.version = version;
        switch (version) {
            case "1.0":
                this.buildPath = '';
                this.sourcePath = '';
                this.skipMetadataFiles = true;
                break;
            case "2.0":
                this.buildPath = '$PROJECTPATH/build';
                this.sourcePath = '$PROJECTPATH/source';
                this.skipMetadataFiles = false;
                break;
            default:
                throw new Error('Invalid version');
        }
    }
}
;
;
export class KeymanDeveloperProjectFile10 {
    id;
    filePath;
    fileVersion;
    path;
    get filename() {
        return this.path.basename(this.filePath);
    }
    get fileType() {
        return KeymanFileTypes.fromFilename(this.filename);
    }
    details; // 1.0 only
    childFiles; // 1.0 only
    constructor(id, filePath, fileVersion, path) {
        this.id = id;
        this.filePath = filePath;
        this.fileVersion = fileVersion;
        this.path = path;
        this.details = {};
        this.childFiles = [];
    }
}
;
export class KeymanDeveloperProjectFile20 {
    filePath;
    path;
    get filename() {
        return this.path.basename(this.filePath);
    }
    get fileType() {
        return KeymanFileTypes.fromFilename(this.filename);
    }
    constructor(filePath, path) {
        this.filePath = filePath;
        this.path = path;
    }
}
;
export class KeymanDeveloperProjectFileDetail_Kps {
    name;
    copyright;
    version;
}
;
export class KeymanDeveloperProjectFileDetail_Kmn {
    name;
    copyright;
    message;
}
;
//# sourceMappingURL=keyman-developer-project.js.map