/*
 * Keyman is copyright (C) SIL Global. MIT License.
 *
 * Created by srl on 2024-09-27
 *
 * Abstraction for XML reading and writing
 */
import { XMLParser, XMLBuilder } from 'fast-xml-parser';
const PARSER_COMMON_OPTIONS = {
    attributeNamePrefix: '$', // causes remapping into $: { … } objects
    htmlEntities: true,
    ignoreAttributes: false,
    ignorePiTags: true,
    numberParseOptions: {
        eNotation: null,
        hex: null,
        leadingZeros: null,
        skipLike: /(?:)/, // parse numbers as strings
    },
    textNodeName: '_',
};
/** map of options for the XML parser */
const PARSER_OPTIONS = {
    'keyboard3': {
        attributeNamePrefix: '@__', // We'll use this to convert attributes to strings and subobjects to arrays, when empty.
        htmlEntities: true,
        ignoreAttributes: false, // We'd like attributes, please
        tagValueProcessor: (_tagName, tagValue /*, jPath, hasAttributes, isLeafNode*/) => {
            // since trimValues: false, we need to zap any element values that would be trimmed.
            // currently, the LDML spec doesn't have any element values, but this
            // future-proofs us a little in that element values are allowed, just trimmed.
            // if we do need elements in the future, we'd check the preserve-space attribute here.
            return tagValue?.trim();
        },
        trimValues: false, // preserve spaces, but see tagValueProcessor
    },
    'keyboardTest3': {
        attributeNamePrefix: '', // avoid @_
        htmlEntities: true,
        ignoreAttributes: false, // We'd like attributes, please
        ignorePiTags: true,
        preserveOrder: true, // Gives us a 'special' format
    },
    'kps': {
        ...PARSER_COMMON_OPTIONS,
    },
    'kpj': {
        ...PARSER_COMMON_OPTIONS,
        attributeNamePrefix: '', // to avoid '@_' prefixes
    },
    'kvks': {
        ...PARSER_COMMON_OPTIONS,
        tagValueProcessor: (_tagName, tagValue, _jPath, _hasAttributes, isLeafNode) => {
            if (!isLeafNode) {
                return tagValue?.trim(); // trimmed value
            }
            else {
                return null; // no change to leaf nodes
            }
        },
        trimValues: false, // preserve spaces
    },
};
const GENERATOR_COMMON_OPTIONS = {
    attributeNamePrefix: '$',
    ignoreAttributes: false,
    format: true,
    textNodeName: '_',
    suppressEmptyNode: true,
};
const GENERATOR_OPTIONS = {
    kvks: {
        ...GENERATOR_COMMON_OPTIONS,
    },
    kpj: {
        ...GENERATOR_COMMON_OPTIONS,
    },
    kps: {
        ...GENERATOR_COMMON_OPTIONS,
    },
};
/** wrapper for XML parsing support */
export class KeymanXMLReader {
    type;
    constructor(type) {
        this.type = type;
    }
    /** move `{ $abc: 4 }` into `{ $: { abc: 4 } }` */
    static fixupDollarAttributes(data) {
        if (typeof data === 'object') {
            if (Array.isArray(data)) {
                return data.map(v => KeymanXMLReader.fixupDollarAttributes(v));
            }
            // object
            const e = [];
            const attrs = [];
            Object.entries(data).forEach(([k, v]) => {
                if (k[0] === '$') {
                    k = k.slice(1);
                    attrs.push([k, KeymanXMLReader.fixupDollarAttributes(v)]);
                }
                else {
                    e.push([k, KeymanXMLReader.fixupDollarAttributes(v)]);
                }
            });
            if (attrs.length) {
                e.push(['$', Object.fromEntries(attrs)]);
            }
            return Object.fromEntries(e);
        }
        else {
            return data;
        }
    }
    /**
     * Requires attribute prefix @__ (double underscore)
     * For attributes, just remove @__ and continue.
     * For objects, replace any empty string "" with an empty object {} */
    static fixupEmptyStringToEmptyObject(data) {
        if (typeof data === 'object') {
            // For arrays of objects, we map "" to {}
            // "" means an empty object
            if (Array.isArray(data)) {
                return data.map(v => {
                    if (v === '') {
                        return {};
                    }
                    else {
                        return KeymanXMLReader.fixupEmptyStringToEmptyObject(v);
                    }
                });
            }
            // otherwise: remove @__ for attributes, remap objects
            const e = [];
            Object.entries(data).forEach(([k, v]) => {
                if (k.startsWith('@__')) {
                    e.push([k.substring(3), KeymanXMLReader.fixupEmptyStringToEmptyObject(v)]);
                }
                else {
                    if (v === '') {
                        e.push([k, {}]);
                    }
                    else {
                        e.push([k, KeymanXMLReader.fixupEmptyStringToEmptyObject(v)]);
                    }
                }
            });
            return Object.fromEntries(e);
        }
        else {
            return data;
        }
    }
    /**
     * Replace:
     * ```json
     * [ { "info": [], ":@": { "abc": "def" } }]
     * ```
     * with:
     * ```json
     * [{"$": { "abc": "def" }, "#name": "info" }]
     * ```
     * see https://github.com/NaturalIntelligence/fast-xml-parser/blob/master/docs/v4/2.XMLparseOptions.md#preserveorder
     * @param data input data
     */
    static fixupPreserveOrder(data) {
        // we need to extract the root name specially
        if (!Array.isArray(data)) {
            throw Error(`Internal Error: XML parser preserveOrder did not yield an array.`);
        }
        if (data.length !== 1) {
            // we ignore comments, so should only have one element
            throw Error(`Internal Error: XML parser preserveOrder did not yield an array of size 1.`);
        }
        // the root element is special, we copy it into a property
        const rootElement = KeymanXMLReader.fixupPreserveOrderObject(data[0]);
        const rootElementName = rootElement['#name'];
        const out = {};
        out[rootElementName] = rootElement;
        return out;
    }
    /** takes an 'object' with a property `:@` containing attrs, and one other property with the object name */
    static fixupPreserveOrderObject(data) {
        const attrs = data[':@'];
        const mainEntry = Object.entries(data).filter(([k, v]) => k !== ':@');
        const [elementName, subItems] = mainEntry[0];
        const out = {};
        if (attrs) {
            out['$'] = attrs;
        }
        if (!elementName) {
            throw Error(`could not find elementName in ${JSON.stringify(mainEntry[0])}`);
        }
        out['#name'] = elementName;
        if (subItems && subItems.length) {
            out['$$'] = subItems.map((subObject) => KeymanXMLReader.fixupPreserveOrderObject(subObject));
            // xml2js duplicated data here, including elements in their 'non-preserved-order' form.
            // we don't read this data, but we're maintaining compatibility here with the read format.
            // example:  emit: […], keystroke:[…]
            for (const o of out['$$']) {
                const subElementName = o['#name'];
                const nonPreservedElements = out[subElementName] = out[subElementName] ?? [];
                const oWithoutName = { ...o };
                delete oWithoutName['#name']; // #name is only there in the preserved-order form.
                nonPreservedElements.push(oWithoutName);
            }
        }
        return out;
    }
    parse(data) {
        const parser = this.parser();
        let result = parser.parse(data, true);
        if (PARSER_OPTIONS[this.type].attributeNamePrefix === '$') {
            result = KeymanXMLReader.fixupDollarAttributes(result);
        }
        else if (PARSER_OPTIONS[this.type].attributeNamePrefix === '@__') {
            result = KeymanXMLReader.fixupEmptyStringToEmptyObject(result);
        }
        else if (PARSER_OPTIONS[this.type].preserveOrder) {
            result = KeymanXMLReader.fixupPreserveOrder(result);
        }
        delete result['?xml'];
        return result;
    }
    parser() {
        let options = PARSER_OPTIONS[this.type];
        if (!options) {
            /* c8 ignore next 1 */
            throw Error(`Internal error: unhandled XML type ${this.type}`);
        }
        options = Object.assign({}, options); // TODO: xml2js likes to mutate the options here. Shallow clone the object.
        return new XMLParser(options);
    }
}
/**
 * Fixed prologue for writing XML
 */
const PROLOGUE = { '?xml': { '$version': '1.0', '$encoding': 'utf-8' } };
/** wrapper for XML generation support */
export class KeymanXMLWriter {
    type;
    static fixDataForWrite(data) {
        if (typeof data === 'object') {
            if (Array.isArray(data)) {
                // just fixup each item of the array
                return data.map(d => KeymanXMLWriter.fixDataForWrite(d));
            }
            // else object
            const e = [];
            Object.entries(data).forEach(([k, v]) => {
                if (k === '$') {
                    /* convert $: { a: 1, b: 2 } to { $a: 1, $b: 2} */
                    Object.entries(v).forEach(([k, v]) => {
                        e.push([`\$${k}`, KeymanXMLWriter.fixDataForWrite(v)]);
                    });
                }
                else {
                    e.push([k, KeymanXMLWriter.fixDataForWrite(v)]);
                }
            });
            // reconstitute with $ elements fixed
            return Object.fromEntries(e);
        }
        else {
            return data; // string or something else
        }
    }
    write(data) {
        const builder = this.builder();
        data = KeymanXMLWriter.fixDataForWrite(data);
        return builder.build({ ...PROLOGUE, ...data });
    }
    constructor(type) {
        this.type = type;
    }
    builder() {
        const options = GENERATOR_OPTIONS[this.type];
        if (!options) {
            /* c8 ignore next 1 */
            throw Error(`Internal error: unhandled XML type ${this.type}`);
        }
        return new XMLBuilder(Object.assign({}, options)); // Shallow clone in case the options are mutated.
    }
}
//# sourceMappingURL=xml-utils.js.map